import subprocess
from typing import Optional
from miyuki.config import FFMPEG_INPUT_FILE
from miyuki.logger import logger


class FFmpegProcessor:
    @staticmethod
    def create_video_from_segments(segment_files: list[str], output_file: str, cover_file: Optional[str] = None) -> None:
        with open(FFMPEG_INPUT_FILE, 'w') as f:
            for file in segment_files:
                f.write(f"file '{file}'\n")
        ffmpeg_command = ['ffmpeg', '-y', '-loglevel', 'error', '-f', 'concat', '-safe', '0', '-i', FFMPEG_INPUT_FILE]
        if cover_file:
            ffmpeg_command.extend(['-i', cover_file, '-map', '0', '-map', '1', '-c', 'copy', '-disposition:v:1', 'attached_pic'])
        else:
            ffmpeg_command.extend(['-c', 'copy'])
        ffmpeg_command.append(output_file)
        try:
            subprocess.run(ffmpeg_command, check=True, stdout=subprocess.DEVNULL)
            logger.info("FFmpeg execution completed.")
        except subprocess.CalledProcessError as e:
            logger.error(f"FFmpeg execution failed: {e}")
            raise



import threading
import sys
import os
import shutil
from typing import Tuple


class ThreadSafeCounter:
    def __init__(self) -> None:
        self._count = 0
        self._lock = threading.Lock()

    def increment_and_get(self) -> int:
        with self._lock:
            self._count += 1
            return self._count

    def get(self) -> int:
        with self._lock:
            return self._count

    def reset(self) -> None:
        with self._lock:
            self._count = 0


def display_progress_bar(max_value: int, file_counter: ThreadSafeCounter) -> None:
    bar_length = 50
    current_value = file_counter.increment_and_get()
    progress = current_value / max_value
    block = int(round(bar_length * progress))
    text = f"\rProgress: [{'#' * block + '-' * (bar_length - block)}] {current_value}/{max_value}"
    sys.stdout.write(text)
    sys.stdout.flush()


def split_integer_into_intervals(integer: int, n: int) -> list[Tuple[int, int]]:
    interval_size = integer // n
    remainder = integer % n
    intervals = [(i * interval_size, (i + 1) * interval_size) for i in range(n)]
    if remainder:
        intervals[-1] = (intervals[-1][0], intervals[-1][1] + remainder)
    return intervals


def find_last_non_empty_line(text: str) -> str:
    lines = text.splitlines()
    for line in reversed(lines):
        if line.strip():
            return line
    raise Exception("Failed to find the last non-empty line in m3u8 playlist.")


def find_closest(arr: list[int], target: int) -> int:
    closest = arr[0]
    min_diff = abs(arr[0] - target)
    for num in arr:
        diff = abs(num - target)
        if diff < min_diff:
            min_diff = diff
            closest = num
    return closest


def delete_all_subfolders(folder_path: str) -> None:
    if not os.path.exists(folder_path):
        return
    for item in os.listdir(folder_path):
        item_path = os.path.join(folder_path, item)
        if os.path.isdir(item_path):
            shutil.rmtree(item_path)



# --- START OF FILE miyuki/url_sources.py ---

from abc import ABC, abstractmethod
import re
from typing import Optional
# from miyuki.http_client import HttpClient # No longer needed
from miyuki.config import HREF_REGEX_PUBLIC_PLAYLIST, HREF_REGEX_NEXT_PAGE, MATCH_UUID_PATTERN
from miyuki.logger import logger
from miyuki.utils import ThreadSafeCounter
from enum import Enum
import cloudscraper # Use cloudscraper
from urllib.parse import urljoin # Needed for SearchSource

class UrlType(Enum):
    SINGLE = 1
    PLAYLIST = 2

class UrlSource(ABC):
    @abstractmethod
    def get_urls(self) -> list[str]:
        pass

    @staticmethod
    def movie_count_log(movie_counter: ThreadSafeCounter, movie_url: str):
        count = movie_counter.increment_and_get() # Get count first
        logger.info(f"Identified Movie URL #{count}: {movie_url}") # Log count and URL

    @staticmethod
    def get_urls_from_list(movie_counter: ThreadSafeCounter, play_list_url: str, limit: Optional[int], scraper: cloudscraper.CloudScraper, cookie=None) -> list[str]: # Accept scraper
        movie_url_list = []
        url = play_list_url
        base_url = "/".join(play_list_url.split("/")[:3]) # Extract base URL like https://missav.ai
        logger.info(f"[get_urls_from_list] Starting playlist processing for URL: {url}, Limit: {limit}, Base: {base_url}")
        page_count = 0
        # Make sure limit comparison works (limit is int | None)
        while url and (limit is None or movie_counter.get() < limit):
            page_count += 1
            current_page_url = url # Store the URL for this iteration
            logger.info(f"[get_urls_from_list] Fetching page {page_count} URL: {current_page_url}")

            try:
                 response = scraper.get(current_page_url, cookies=cookie, timeout=20) # Use scraper
                 response.raise_for_status() # Check for HTTP errors (4xx, 5xx)
                 html_source = response.text # Use .text for decoded string
                 logger.debug(f"[get_urls_from_list] Successfully fetched HTML for: {current_page_url}")
                 # Save the HTML for debugging playlist issues if needed
                 # with open(f"debug_playlist_page_{page_count}.html", "w", encoding="utf-8") as pf: pf.write(html_source)
            except cloudscraper.exceptions.CloudflareException as e:
                 logger.error(f"[get_urls_from_list] Cloudflare challenge failed for {current_page_url}: {e}")
                 break # Stop processing this playlist if blocked
            except Exception as e:
                 logger.error(f"[get_urls_from_list] Failed to fetch HTML for page URL {current_page_url}: {e}")
                 break # Stop processing this playlist on other errors

            # --- Find Movie Links ---
            # Check if HREF_REGEX_PUBLIC_PLAYLIST is still valid for the site structure
            # Example: '<a href="([^"]+)" alt="[^"]*">' might be too specific if 'alt' is missing
            # A potentially more robust pattern might just look for links within specific list items:
            # Example: '<div class="card-body">\s*<a href="([^"]+)"' (NEEDS INSPECTION of playlist HTML)
            movie_url_matches = re.findall(HREF_REGEX_PUBLIC_PLAYLIST, html_source)
            logger.info(f"[get_urls_from_list] Found {len(movie_url_matches)} potential movie links using pattern '{HREF_REGEX_PUBLIC_PLAYLIST}' on page {page_count}.")
            logger.debug(f"[get_urls_from_list] Links found: {movie_url_matches}")

            temp_url_list = list(set(movie_url_matches)) # Get unique URLs from this page
            logger.info(f"[get_urls_from_list] Found {len(temp_url_list)} unique potential movie links on page {page_count}.")

            added_count_this_page = 0
            for movie_url_relative in temp_url_list:
                # Check limit before adding and logging
                if limit is not None and movie_counter.get() >= limit:
                     logger.info(f"[get_urls_from_list] Limit ({limit}) reached. Stopping collection.")
                     return movie_url_list # Return immediately

                # Make URL absolute using the base URL
                absolute_movie_url = urljoin(base_url, movie_url_relative.strip())

                movie_url_list.append(absolute_movie_url)
                UrlSource.movie_count_log(movie_counter, absolute_movie_url) # Log happens here with increment
                added_count_this_page += 1

            logger.info(f"[get_urls_from_list] Added {added_count_this_page} new URLs from page {page_count}. Total collected so far: {len(movie_url_list)}")

            # Check limit again after processing the page, before finding next page
            if limit is not None and movie_counter.get() >= limit:
                 logger.info(f"[get_urls_from_list] Limit ({limit}) reached after processing page. Stopping collection.")
                 return movie_url_list # Return immediately

            # --- Find Next Page Link ---
            # Check if HREF_REGEX_NEXT_PAGE is still valid
            next_page_matches = re.findall(HREF_REGEX_NEXT_PAGE, html_source)
            if next_page_matches:
                next_page_relative_url = next_page_matches[0].replace('&', '&')
                # Make the next page URL absolute relative to the *current page's* full URL
                # OR assume it's relative to the base_url - needs checking on the site
                # Let's assume relative to base_url for now
                url = urljoin(base_url, next_page_relative_url)
                logger.info(f"[get_urls_from_list] Found next page link: {url}")
            else:
                logger.info(f"[get_urls_from_list] No next page link found on page {page_count}. End of playlist.")
                url = None # Stop the loop
        logger.info(f"[get_urls_from_list] Finished playlist processing for {play_list_url}. Total URLs collected: {len(movie_url_list)}")
        return movie_url_list

class SingleUrlSource(UrlSource):
    # No changes needed here, doesn't make HTTP requests directly
    def __init__(self, movie_counter: ThreadSafeCounter, url: str, limit: Optional[int]):
        self.movie_counter = movie_counter
        self.url = url
        self.limit = limit
        logger.info(f"[SingleUrlSource.__init__] Initialized for URL: {self.url}, Limit: {self.limit}")

    def get_urls(self) -> list[str]:
        logger.info(f"[SingleUrlSource.get_urls] Checking URL: {self.url}")
        if self.limit is not None and self.movie_counter.get() >= self.limit: # Check limit before adding
            logger.info(f"[SingleUrlSource.get_urls] Limit ({self.limit}) reached. Skipping URL: {self.url}")
            return []
        else:
            UrlSource.movie_count_log(self.movie_counter, self.url)
            logger.info(f"[SingleUrlSource.get_urls] Returning URL: {self.url}")
            return [self.url]

class PlaylistSource(UrlSource):
    # Accept scraper in __init__
    def __init__(self, movie_counter: ThreadSafeCounter, playlist_url: str, limit: Optional[int], scraper: cloudscraper.CloudScraper):
        self.movie_counter = movie_counter
        self.playlist_url = playlist_url
        self.limit = limit
        self.scraper = scraper # Store the passed scraper
        logger.info(f"[PlaylistSource.__init__] Initialized for Playlist URL: {self.playlist_url}, Limit: {self.limit}")

    def get_urls(self) -> list[str]:
        logger.info(f"[PlaylistSource.get_urls] Starting processing for playlist: {self.playlist_url}")
        # Pass the stored scraper
        result_urls = UrlSource.get_urls_from_list(movie_counter=self.movie_counter, play_list_url=self.playlist_url, limit=self.limit, scraper=self.scraper, cookie=None)
        logger.info(f"[PlaylistSource.get_urls] Finished processing for {self.playlist_url}. Found {len(result_urls)} URLs.")
        return result_urls

class AutoUrlSource(UrlSource):
     # Accept scraper in __init__
    def __init__(self, movie_counter: ThreadSafeCounter, auto_urls: list[str], limit: Optional[str], scraper: cloudscraper.CloudScraper):
        self.movie_counter = movie_counter
        self.auto_urls = auto_urls
        # Convert limit to int here, handling None
        self.limit = int(limit) if limit else None
        self.scraper = scraper # Store the passed scraper
        logger.info(f"[AutoUrlSource.__init__] Initialized with URLs: {self.auto_urls}, Limit: {self.limit}")

    def get_urls(self) -> list[str]:
        movie_url_list = []
        logger.info(f"[AutoUrlSource.get_urls] Starting processing for {len(self.auto_urls)} provided URLs.")

        for url in self.auto_urls:
            logger.info(f"[AutoUrlSource.get_urls] Processing input URL: {url}")

            # Check limit before even determining type
            if self.limit is not None and self.movie_counter.get() >= self.limit:
                logger.info(f"[AutoUrlSource.get_urls] Limit ({self.limit}) reached. Skipping further URL processing.")
                break # Stop processing more URLs from the input list

            # Pass scraper when determining type
            url_type : Optional[UrlType] = self._determine_url_type(url, self.scraper)
            logger.info(f"[AutoUrlSource.get_urls] Determined URL type for {url}: {url_type}")

            sub_source_urls = []
            if url_type == UrlType.SINGLE:
                logger.info(f"[AutoUrlSource.get_urls] Treating {url} as SINGLE. Creating SingleUrlSource.")
                # SingleUrlSource doesn't need scraper passed to it
                single_url_source = SingleUrlSource(movie_counter=self.movie_counter, url=url, limit=self.limit)
                sub_source_urls = single_url_source.get_urls()
                logger.info(f"[AutoUrlSource.get_urls] SingleUrlSource for {url} returned {len(sub_source_urls)} URL(s).")
            elif url_type == UrlType.PLAYLIST:
                logger.info(f"[AutoUrlSource.get_urls] Treating {url} as PLAYLIST. Creating PlaylistSource.")
                # Pass the scraper to PlaylistSource
                playlist_source = PlaylistSource(movie_counter=self.movie_counter, playlist_url=url, limit=self.limit, scraper=self.scraper)
                sub_source_urls = playlist_source.get_urls()
                logger.info(f"[AutoUrlSource.get_urls] PlaylistSource for {url} returned {len(sub_source_urls)} URL(s).")
            else:
                logger.warning(f"[AutoUrlSource.get_urls] Could not determine type for URL: {url}. Skipping.")

            if sub_source_urls:
                 # Ensure URLs are unique if accumulating from multiple sources
                 # Although PlaylistSource already handles uniqueness per page
                 movie_url_list.extend(u for u in sub_source_urls if u not in movie_url_list)
                 logger.info(f"[AutoUrlSource.get_urls] Added {len(sub_source_urls)} unique URL(s) to main list. Total now: {len(movie_url_list)}")

        logger.info(f"[AutoUrlSource.get_urls] Finished processing all input URLs. Final list size: {len(movie_url_list)}")
        return movie_url_list

    # Modify _determine_url_type to accept scraper
    def _determine_url_type(self, url: str, scraper: cloudscraper.CloudScraper) -> Optional[UrlType]:
        logger.info(f"[AutoUrlSource._determine_url_type] Checking URL: {url}")
        try:
            # Pass scraper to _is_movie_url
            is_movie = self._is_movie_url(url, scraper)
            logger.info(f"[AutoUrlSource._determine_url_type] Result of _is_movie_url for {url}: {is_movie}")
            if is_movie:
                logger.info(f"[AutoUrlSource._determine_url_type] Determined type: SINGLE for {url}")
                return UrlType.SINGLE
            else:
                # Add a check here? Could try fetching as playlist and see if it finds links?
                # For now, stick to the original logic: if not movie -> playlist
                logger.info(f"[AutoUrlSource._determine_url_type] Determined type: PLAYLIST for {url}")
                return UrlType.PLAYLIST
        except Exception as e:
             logger.error(f"[AutoUrlSource._determine_url_type] Error determining type for {url}: {e}", exc_info=True)
             return None # Indicate failure

    # Modify _is_movie_url to accept scraper
    def _is_movie_url(self, url: str, scraper: cloudscraper.CloudScraper) -> bool:
        logger.info(f"[AutoUrlSource._is_movie_url] Attempting to fetch HTML for URL type check: {url}")
        try:
             response = scraper.get(url, timeout=20)
             response.raise_for_status()
             html_str = response.text # Use text
             logger.info(f"[AutoUrlSource._is_movie_url] Successfully fetched and decoded HTML for {url}.")
             # Save the HTML we actually get AFTER Cloudflare attempt
        except cloudscraper.exceptions.CloudflareException as e:
             logger.error(f"[AutoUrlSource._is_movie_url] Cloudflare challenge failed for {url}: {e}")
             # Save the challenge page content if possible
             try:
                  with open("debug_cloudflare_challenge.html", "w", encoding="utf-8") as cf_file:
                       cf_file.write(e.response.text)
                  logger.info("Saved Cloudflare challenge page to 'debug_cloudflare_challenge.html'")
             except Exception as save_e:
                  logger.error(f"Could not save Cloudflare challenge page: {save_e}")
             return False # Treat as non-movie if challenge fails
        except Exception as e:
             logger.warning(f"[AutoUrlSource._is_movie_url] Failed to fetch HTML for {url}: {e}. Cannot determine if it's a movie URL.")
             return False

        # Now try the regex on html_str
        match = re.search(MATCH_UUID_PATTERN, html_str, re.DOTALL) # Use the pattern from config

        if not match:
             logger.info(f"[AutoUrlSource._is_movie_url] Regex pattern to find eval block did NOT match for {url}. Assuming not a movie URL.")
             return False
        else:
             # We don't need the group here, just that it matched.
             logger.info(f"[AutoUrlSource._is_movie_url] Regex pattern to find eval block matched for {url}. Assuming it IS a movie URL.")
             return True

# --- Modify AuthSource ---
class AuthSource(UrlSource):
     # Accept scraper in __init__
    def __init__(self, movie_counter: ThreadSafeCounter, username: str, password: str, limit: Optional[str], scraper: cloudscraper.CloudScraper):
        self.movie_counter = movie_counter
        self.scraper = scraper # Store scraper
        self.limit = int(limit) if limit else None
        # Login needs to use the scraper now
        self.cookie = self._login(username, password) # Store cookies returned by login if needed

    def _login(self, username: str, password: str) -> Optional[dict]: # Return Optional dict
        logger.info(f"Attempting login for user: {username}")
        try:
            # Use scraper.post
            response = self.scraper.post('https://missav.ai/api/login', data={'email': username, 'password': password}, timeout=20)
            logger.debug(f"Login response status: {response.status_code}")
            logger.debug(f"Login response headers: {response.headers}")
            logger.debug(f"Login response content snippet: {response.text[:200]}") # Log snippet
            response.raise_for_status() # Check for HTTP errors

            if response.status_code == 200:
                # Check if login was actually successful based on response content or cookies set
                # This part is tricky without knowing the exact API response for success/fail
                # Let's assume for now it worked and cloudscraper handles session cookies
                logger.info(f"Login request seems successful (Status 200). Relying on cloudscraper session cookies.")
                return self.scraper.cookies.get_dict() # Return current session cookies
            else:
                # This part might not be reached if raise_for_status catches the error
                logger.error(f"Login request failed with status: {response.status_code}")
                exit(MAGIC_NUMBER)
        except cloudscraper.exceptions.CloudflareException as e:
             logger.error(f"Cloudflare challenge failed during login: {e}")
             exit(MAGIC_NUMBER)
        except Exception as e:
             logger.error(f"Login failed: Check network or account info. Error: {e}")
             exit(MAGIC_NUMBER)
        # Added return None for clarity if exit doesn't happen
        return None

    def get_urls(self) -> list[str]:
        url = 'https://missav.ai/saved'
        logger.info(f"Fetching saved URLs from: {url}")
        if self.cookie is None:
             logger.error("Login failed previously, cannot fetch saved URLs.")
             return []
        # Pass the scraper to get_urls_from_list.
        # Cookies are managed by the scraper's session after successful login.
        return UrlSource.get_urls_from_list(movie_counter=self.movie_counter, play_list_url=url, limit=self.limit, scraper=self.scraper, cookie=None)

# --- Modify SearchSource ---
class SearchSource(UrlSource):
     # Accept scraper in __init__
    def __init__(self, movie_counter: ThreadSafeCounter, key: str, scraper: cloudscraper.CloudScraper, cookies: Optional[dict] = None): # Added scraper
        self.movie_counter = movie_counter
        self.key = key
        self.scraper = scraper # Store scraper
        self.cookies = cookies # Keep cookies if needed

    def get_urls(self) -> list[str]:
        search_url = f"https://missav.ai/search/{self.key}"
        # Define base URL for resolving relative links found
        base_url = "https://missav.ai"
        # Escape key for safety in regex and use the public playlist regex
        search_regex = HREF_REGEX_PUBLIC_PLAYLIST # Use the same regex as playlists
        logger.info(f"Searching for key '{self.key}' at URL: {search_url}")

        try:
            # Use scraper.get
            response = self.scraper.get(search_url, cookies=self.cookies, timeout=20)
            response.raise_for_status()
            html_source = response.text # Use .text
            logger.debug(f"Successfully fetched search results page for key '{self.key}'")
            # Save for debugging if needed
            # with open(f"debug_search_{self.key}.html", "w", encoding="utf-8") as sf: sf.write(html_source)
        except cloudscraper.exceptions.CloudflareException as e:
             logger.error(f"Cloudflare challenge failed during search for key '{self.key}': {e}")
             return []
        except Exception as e:
             logger.error(f"Search failed for key '{self.key}': {e}")
             return []

        # Find all links matching the playlist pattern
        movie_url_matches = re.findall(search_regex, html_source)
        logger.info(f"Found {len(movie_url_matches)} potential matches on search results page.")

        # Filter results to find the specific key if needed, or just take the first one
        # The original regex r'<a href="([^"]+)" alt="' + self.key + '" >' tried to match the alt tag.
        # If HREF_REGEX_PUBLIC_PLAYLIST doesn't include alt, we might need a different approach
        # or iterate through matches and check their text/alt tags if necessary.
        # For now, let's assume the first match is the desired one if any are found.

        if movie_url_matches:
             # Take the first unique relative URL found
             first_relative_url = list(set(movie_url_matches))[0]
             # Make it absolute
             absolute_url = urljoin(base_url, first_relative_url.strip())

             logger.info(f"Search for '{self.key}' successful. Using first found URL: {absolute_url}")
             UrlSource.movie_count_log(self.movie_counter, absolute_url)
             return [absolute_url] # Return list with one URL
        else:
             logger.error(f"Search key '{self.key}' did not yield any matching links on the results page using pattern '{search_regex}'.")
             return []

# --- Modify FileSource ---
class FileSource(UrlSource):
     # Accept scraper in __init__
    def __init__(self, movie_counter: ThreadSafeCounter, file_path: str, limit: Optional[str], scraper: cloudscraper.CloudScraper): # Added scraper
        self.movie_counter = movie_counter
        self.file_path = file_path
        self.limit = int(limit) if limit else None
        self.scraper = scraper # Store scraper
        logger.info(f"Initializing FileSource with path: {self.file_path}")

    def get_urls(self) -> list[str]:
        logger.info(f"Reading URLs from file: {self.file_path}")
        try:
             with open(self.file_path, 'r', encoding='utf-8') as f:
                # Read lines, strip whitespace, filter out empty lines
                urls = [line.strip() for line in f if line.strip()]
             logger.info(f"Read {len(urls)} URLs from file.")
        except FileNotFoundError:
             logger.error(f"File not found: {self.file_path}")
             return []
        except Exception as e:
             logger.error(f"Error reading file {self.file_path}: {e}")
             return []

        if not urls:
             logger.warning(f"No valid URLs found in file: {self.file_path}")
             return []

        # Pass the scraper to AutoUrlSource instance created here
        # Ensure limit is passed correctly (as string or None)
        limit_str = str(self.limit) if self.limit is not None else None
        logger.info(f"Creating AutoUrlSource for file content with limit: {limit_str}")
        auto_url_source = AutoUrlSource(movie_counter=self.movie_counter, auto_urls=urls, limit=limit_str, scraper=self.scraper)
        return auto_url_source.get_urls()

# --- END OF FILE miyuki/url_sources.py ---


import logging
import sys

logger = logging.getLogger('miyuki-logger')
logger.setLevel(logging.DEBUG)

file_handler = logging.FileHandler('miyuki.log', encoding='utf-8')
file_handler.setLevel(logging.DEBUG)

console_handler = logging.StreamHandler(sys.stdout)
console_handler.setLevel(logging.DEBUG)

formatter = logging.Formatter('Miyuki - %(asctime)s - %(levelname)s - %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
file_handler.setFormatter(formatter)
console_handler.setFormatter(formatter)

logger.addHandler(file_handler)
logger.addHandler(console_handler)



import os
import re
from typing import Optional, Tuple
import threading
from miyuki.config import ( # Make sure to import all needed config vars
    MOVIE_SAVE_PATH_ROOT, MATCH_UUID_PATTERN, MATCH_TITLE_PATTERN, COVER_URL_PREFIX,
    TMP_HTML_FILE, RESOLUTION_PATTERN, VIDEO_M3U8_PREFIX, VIDEO_PLAYLIST_SUFFIX
)
from miyuki.logger import logger
from miyuki.utils import ThreadSafeCounter, display_progress_bar, split_integer_into_intervals, find_last_non_empty_line, find_closest
from miyuki.ffmpeg_processor import FFmpegProcessor
import cloudscraper


class VideoDownloader:
    def __init__(self, url: str, scraper: cloudscraper.CloudScraper, options: dict):
        self.url = url
        self.scraper = scraper # Store scraper instance
        self.movie_name = url.split('/')[-1]
        self.movie_folder = os.path.join(MOVIE_SAVE_PATH_ROOT, self.movie_name)
        self.options = options
        self.uuid = None
        self.title = None
        self.final_file_name = None
        self.counter = ThreadSafeCounter()

    def _fetch_metadata(self) -> bool:
        try:
            # Use scraper.get().text for HTML processing
            response = self.scraper.get(self.url, timeout=20)
            response.raise_for_status()
            html = response.text # Use .text
        except cloudscraper.exceptions.CloudflareException as e:
             logger.error(f"Cloudflare challenge failed fetching metadata for {self.url}: {e}")
             return False
        except Exception as e:
             logger.error(f"Failed to fetch HTML for {self.url}: {e}")
             return False

        # No decode needed

        with open(TMP_HTML_FILE, 'w', encoding='utf-8') as file:
            file.write(html)

        # Use the NEW pattern, and add re.DOTALL in case the JS is multiline
        match = re.search(MATCH_UUID_PATTERN, html, re.DOTALL)
        if not match:
            # Add a log if the *old* pattern is found, indicating partial regression or A/B testing by the site
            old_pattern_match = re.search(r'm3u8\|([a-f0-9\|]+)\|com\|surrit\|https\|video', html)
            if old_pattern_match:
                logger.warning("Found OLD UUID pattern structure. Website might be inconsistent. Trying to parse new structure failed.")
            else:
                logger.error("Failed to match new UUID pattern structure (eval block).")
            return False

        # --- Logic to process the captured split string ---
        # *** IMPORTANT: Adjust group index based on the final regex ***
        # If using: r"eval\(function\(p,a,c,k,e,d\)\{.*?\}\((['\"])(.*?)\1\s*\.\s*split\(\s*'\|'\s*\)"
        # The content is in Group 2
        split_string_content = match.group(2) # <-- USE GROUP 2
        logger.info("Successfully matched eval split string content.")
        logger.debug(f"Split string content raw: '{split_string_content}'")

        parts = split_string_content.split('|')
        logger.debug(f"Split parts array (length {len(parts)}): {parts}")

        required_indices = [1, 2, 3, 4, 5]
        if not all(i < len(parts) for i in required_indices):
             logger.error(f"Split parts array does not contain all expected indices [1-5] for UUID. Cannot construct UUID.")
             logger.debug(f"Available parts: {parts}")
             return False

        try:
             self.uuid = f"{parts[5]}-{parts[4]}-{parts[3]}-{parts[2]}-{parts[1]}"
             logger.info(f"Constructed UUID successfully: {self.uuid}")
        except IndexError:
             logger.error(f"Error accessing expected indices [1-5] in split parts array during UUID construction.")
             logger.debug(f"Available parts: {parts}")
             return False

        # --- Title extraction remains the same ---
        title_match = re.search(MATCH_TITLE_PATTERN, html)
        if title_match:
            illegal_chars = r'[<>:"/\\|?*\s]+' # Use regex for illegal chars + whitespace
            origin_title = title_match.group(1).strip() # Added strip()
            # Replace sequences of illegal chars/spaces with a single underscore
            safe_title = re.sub(illegal_chars, '_', origin_title)
            # Remove leading/trailing underscores that might result
            safe_title = safe_title.strip('_')

            if "uncensored" in self.url.lower(): # Use lower() for case-insensitivity
                safe_title += "_uncensored"
            self.title = safe_title
            logger.info(f"Found title: '{origin_title}', Sanitized: '{self.title}'")
        else:
             logger.warning(f"Could not extract title for {self.url}")
             self.title = None # Ensure title is None if not found
        # --- End of Title extraction ---

        return True

    def _download_cover(self) -> None:
        if not self.options.get('cover_action'): return
        cover_url = f"{COVER_URL_PREFIX}{self.movie_name}/cover-n.jpg"
        try:
             # Use scraper.get().content for image bytes
             response = self.scraper.get(cover_url, timeout=15)
             response.raise_for_status()
             cover_content = response.content # Use .content
        except Exception as e:
             logger.error(f"Failed to download cover for {self.movie_name}: {e}")
             return # Don't proceed if download fails

        cover_path = os.path.join(MOVIE_SAVE_PATH_ROOT, f"{self.movie_name}-cover.jpg")
        try:
             with open(cover_path, 'wb') as f:
                  f.write(cover_content)
             logger.info(f"Cover downloaded successfully to {cover_path}")
        except IOError as e:
             logger.error(f"Failed to save cover to {cover_path}: {e}")

    def _get_final_quality_and_resolution(self, playlist: str) -> Tuple[Optional[str], Optional[str]]:
         # This function takes playlist content as string, no scraper needed here directly
        matches = re.findall(RESOLUTION_PATTERN, playlist)
        quality_map = {height: width for width, height in matches}
        quality_list = list(quality_map.keys())
        if not quality_list:
            logger.error("No resolutions found in playlist.")
            return None, None
        quality = self.options.get('quality')
        if quality is None:
            final_quality = quality_list[-1] + 'p'
            resolution_url = find_last_non_empty_line(playlist)
        else:
            target = int(quality)
            closest_height = find_closest([int(h) for h in quality_list], target)
            final_quality = str(closest_height) + 'p'
            url_type_x = f"{quality_map[str(closest_height)]}x{closest_height}/video.m3u8"
            url_type_p = f"{closest_height}p/video.m3u8"
            resolution_url = url_type_x if url_type_x in playlist else url_type_p if url_type_p in playlist else find_last_non_empty_line(playlist)
        return final_quality, resolution_url

    def _thread_task(self, start: int, end: int, uuid: str, resolution: str, video_offset_max: int) -> None:
        # Create a new scraper instance for each thread to potentially avoid session conflicts
        # Although maybe reusing the main one is ok? Let's try separate first.
        thread_scraper = cloudscraper.create_scraper(delay=5) # Create scraper for this thread
        
        for i in range(start, end):
            url = f"https://surrit.com/{uuid}/{resolution}/video{i}.jpeg"
            try:
                 # Use thread_scraper.get().content
                 retry_count = self.options.get('retry', 5)
                 delay = self.options.get('delay', 2)
                 timeout = self.options.get('timeout', 10)
                 content = None
                 for attempt in range(retry_count):
                      try:
                           response = thread_scraper.get(url, timeout=timeout)
                           response.raise_for_status()
                           content = response.content # Use .content
                           break # Success, exit retry loop
                      except Exception as thread_e:
                           logger.warning(f"Segment {i} failed (attempt {attempt + 1}/{retry_count}): {thread_e}. Retrying in {delay}s...")
                           time.sleep(delay) # Use time.sleep

                 if content:
                      file_path = os.path.join(self.movie_folder, f"video{i}.jpeg")
                      with open(file_path, 'wb') as f:
                            f.write(content)
                      display_progress_bar(video_offset_max + 1, self.counter)
                 else:
                      logger.error(f"Failed to download segment {i} for {self.movie_name} after {retry_count} retries.")

            except Exception as e: # Catch broader exceptions during segment download
                 logger.error(f"Unexpected error downloading segment {i} for {self.movie_name}: {e}")

    def _download_segments(self, uuid: str, resolution: str, video_offset_max: int) -> None:
        if not self.options.get('download_action'):
            return
        intervals = split_integer_into_intervals(video_offset_max + 1, self.options.get('num_threads', os.cpu_count()))
        self.counter.reset()
        threads = []
        for start, end in intervals:
            thread = threading.Thread(target=self._thread_task, args=(start, end, uuid, resolution, video_offset_max))
            threads.append(thread)
            thread.start()
        for thread in threads:
            thread.join()
        self.counter.reset()

    def _check_integrity(self, video_offset_max: int) -> None:
        downloaded_files = len([f for f in os.listdir(self.movie_folder) if f.endswith('.jpeg')])
        total_files = video_offset_max + 1
        integrity = downloaded_files / total_files
        print()
        logger.info(f"File integrity for {self.movie_name}: {integrity:.2%} ({downloaded_files}/{total_files} files)")

    def _assemble_video(self, video_offset_max: int) -> None:
        if not self.options.get('write_action'):
            return
        self.final_file_name = f"{self.movie_name}_{self.final_quality}"
        output_file = os.path.join(MOVIE_SAVE_PATH_ROOT, f"{self.final_file_name}.mp4")
        if self.options.get('ffmpeg_action'):
            segment_files = [os.path.join(self.movie_folder, f"video{i}.jpeg") for i in range(video_offset_max + 1) if os.path.exists(os.path.join(self.movie_folder, f"video{i}.jpeg"))]
            cover_file = os.path.join(MOVIE_SAVE_PATH_ROOT, f"{self.movie_name}-cover.jpg") if self.options.get('cover_as_preview') and os.path.exists(os.path.join(MOVIE_SAVE_PATH_ROOT, f"{self.movie_name}-cover.jpg")) else None
            FFmpegProcessor.create_video_from_segments(segment_files, output_file, cover_file)
        else:
            with open(output_file, 'wb') as outfile:
                for i in range(video_offset_max + 1):
                    file_path = os.path.join(self.movie_folder, f"video{i}.jpeg")
                    if os.path.exists(file_path):
                        with open(file_path, 'rb') as infile:
                            outfile.write(infile.read())
        if self.options.get('title_action') and self.title:
            os.rename(output_file, os.path.join(MOVIE_SAVE_PATH_ROOT, f"{self.title}.mp4"))

    def download(self) -> None:
        if not self._fetch_metadata(): # Uses scraper
            return

        # --- Playlist fetching ---
        playlist_url = f"{VIDEO_M3U8_PREFIX}{self.uuid}{VIDEO_PLAYLIST_SUFFIX}"
        logger.info(f"Fetching main playlist: {playlist_url}")
        try:
             # Use scraper.get().text
             response = self.scraper.get(playlist_url, timeout=15)
             response.raise_for_status()
             playlist = response.text # Use .text
             logger.info("Successfully fetched main playlist.")
        except Exception as e:
             logger.error(f"Failed to fetch playlist {playlist_url}: {e}")
             return

        # No decode needed

        self.final_quality, resolution_url = self._get_final_quality_and_resolution(playlist) # Parses playlist string
        if not self.final_quality or not resolution_url: # Check both
             logger.error("Could not determine final quality or resolution URL.")
             return

        # --- Video m3u8 fetching ---
        video_m3u8_url = f"{VIDEO_M3U8_PREFIX}{self.uuid}/{resolution_url}"
        logger.info(f"Fetching video m3u8: {video_m3u8_url}")
        try:
             # Use scraper.get().text
             response = self.scraper.get(video_m3u8_url, timeout=15)
             response.raise_for_status()
             video_m3u8 = response.text # Use .text
             logger.info("Successfully fetched video m3u8.")
        except Exception as e:
             logger.error(f"Failed to fetch video m3u8 {video_m3u8_url}: {e}")
             return

        # No decode needed

        # --- Segment Offset Parsing ---
        try:
             # Make parsing more robust, find last line ending with .jpeg
             video_offset_max = -1
             lines = video_m3u8.strip().splitlines()
             for line in reversed(lines):
                  line = line.strip()
                  if line.endswith('.jpeg'):
                       match = re.search(r'video(\d+)\.jpeg', line)
                       if match:
                            video_offset_max = int(match.group(1))
                            logger.info(f"Determined max video segment index: {video_offset_max}")
                            break
             if video_offset_max == -1:
                  logger.error("Could not find segment index in video m3u8.")
                  return

        except (IndexError, ValueError, AttributeError) as e:
             logger.error(f"Failed to parse video segment index from m3u8: {e}")
             logger.debug(f"Video m3u8 content:\n{video_m3u8}")
             return

        if not os.path.exists(self.movie_folder):
             os.makedirs(self.movie_folder)

        self._download_cover() # Uses scraper
        self._download_segments(self.uuid, resolution_url.split('/')[0], video_offset_max) # Uses scraper via _thread_task
        self._check_integrity(video_offset_max)
        self._assemble_video(video_offset_max) # Does not use scraper

# Need to import time for sleep in _thread_task
import time



RECORD_FILE = 'downloaded_urls_miyuki.txt'
FFMPEG_INPUT_FILE = 'ffmpeg_input_miyuki.txt'
TMP_HTML_FILE = 'tmp_movie_miyuki.html'
MOVIE_SAVE_PATH_ROOT = 'movies_folder_miyuki'
COVER_URL_PREFIX = 'https://fourhoi.com/'
VIDEO_M3U8_PREFIX = 'https://surrit.com/'
VIDEO_PLAYLIST_SUFFIX = '/playlist.m3u8'
HREF_REGEX_MOVIE_COLLECTION = r'<a class="text-secondary group-hover:text-primary" href="([^"]+)" alt="'
HREF_REGEX_PUBLIC_PLAYLIST = r'<a href="([^"]+)" alt="'
HREF_REGEX_NEXT_PAGE = r'<a href="([^"]+)" rel="next"'
MATCH_UUID_PATTERN = r"eval\(function\(p,a,c,k,e,d\)\{.*?\}\((['\"])(.*?)\1\s*\.\s*split\(\s*'\|'\s*\)"
MATCH_TITLE_PATTERN = r'<title>([^"]+)</title>'
RESOLUTION_PATTERN = r'RESOLUTION=(\d+)x(\d+)'
MAGIC_NUMBER = 114514
RETRY = 5
DELAY = 2
TIMEOUT = 10
HEADERS = {
    'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36',
}
